/***********************************************************************************
*  Sxipper
*  Copyright (c) 2006-2008 Sxip Identity Corporation
*  http://www.sxip.com
*
*  Patents pending.
* 
***********************************************************************************/ 
/*
 * Sxipper - Storage XPCOM Component
 * 
 * Developing notes: change to this file requires removing xpti.dat and compreg.dat
 * from the dev profile directory. 
 */


/***********************************************************
constants
***********************************************************/
const CI = Components.interfaces;

// UUID uniquely identifying our component
const CLASS_ID = Components.ID("{45bb47ca-7b3e-11dc-8314-0800200c9a66}");

// description
const CLASS_NAME = "Sxipper - Data Storage XPCOM Component";

// textual unique identifier
const CONTRACT_ID = "@sxip.com/sxipper/datastore;1";

var smJslib = {};
function smInclude(url) {
	var scriptLoader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"].
		getService(Components.interfaces.mozIJSSubScriptLoader);
	scriptLoader.loadSubScript(url);
}

// bootstrap load jslib
Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
	.getService(Components.interfaces.mozIJSSubScriptLoader)
	.loadSubScript("chrome://sxipper/content/jslib/jslib.js", smJslib);

// load everything else using jslib include
smJslib.include(smJslib.JS_LIB_PATH+"debug/debug.js");
smJslib.include(smJslib.JS_LIB_PATH+"io/file.js");
smJslib.include(smJslib.JS_LIB_PATH+"io/fileUtils.js");
smJslib.include(smJslib.JS_LIB_PATH+"io/dirUtils.js");

smInclude("chrome://sxipper/content/constants.js");
smInclude("chrome://sxipper/content/utils.js");
smInclude("chrome://sxipper/content/promptService.js");
smInclude("chrome://sxipper/content/config.js");
//smInclude("chrome://sxipper/content/message.js");
//smInclude("chrome://sxipper/content/schema.js");
smInclude("chrome://sxipper/content/group-schema.js");
//smInclude("chrome://sxipper/content/model.js");

	
/***********************************************************
 class definition

 insert application wide functions/variables below

***********************************************************/

//class constructor
// the main data model has no name
function smIDataStore(modelName, isLegacyImporter) {
	this.wrappedJSObject = this;
	this.modelName = modelName;
	this.storagePrefix = modelName;
	this._storagePrefixSingletons = {};
	this.isLegacyImporter = isLegacyImporter? true: false;
	this.init();
};

// class definition
smIDataStore.prototype = {

	// define the function we want to expose in our interface
	hello: function() {
		return "XPCOM.smDataStore.hello: Hello World!";
	},

    init: function() {
    	try {
			smUtils.debug("XPCOM.smDataStore.init: initializing store");
			
			// Init internal PKCS#11 token if necessary (fresh profiles ask for master password)
			try {
				var tokenDB = Components.classes["@mozilla.org/security/pk11tokendb;1"]
					.getService(Components.interfaces.nsIPK11TokenDB);
				var token = tokenDB.getInternalKeyToken();
				if (token.needsUserInit) {
					smLog('Init "master password"');
					token.initPassword("");
				}
			} catch (e) {
				smLog('Error while initializing PKCS#11 token: '+e);
			}
			
			// TODO: find a cleaner way to create nested objects
			var filePrototype = function() { };
			filePrototype.prototype = this._filePrototype;
			this.file = new filePrototype();

			var ledgerPrototype = function() { };
			ledgerPrototype.prototype = this._ledgerPrototype;
			this.ledger = new ledgerPrototype();

			var personalPrototype = function() { };
			personalPrototype.prototype = this._personalPrototype;
			this.personal = new personalPrototype();
			
			var passwordManagerPrototype = function() { };
			passwordManagerPrototype.prototype = this._passwordManagerPrototype;
			this.passwordManager = new passwordManagerPrototype();
			
			// allow inner objects to access datastore
			//this.JSON.dataStore = this;
			this.file.dataStore = this;
			this.ledger.dataStore = this;
			this.personal.dataStore = this;
			this.passwordManager.dataStore = this;

			// initialize them
			this.ledger.init();
			this.personal.init();
			this.passwordManager.init();
			
			this.sxipperService = Components.classes["@sxip.com/sxipper/application;1"].getService().wrappedJSObject;
			this.userManager = this.sxipperService.getUserManager();
		} catch (e) {
			dump("XPCOM.smDataStore.init: : error in init(): " + e + "\n" + e.stack);
		}
	},
	
	// creates a new data store object named modelNamed and Prefix all the paths with the modelName
	// useful unit testing
	//  modelName: only alphanumeric characters are accepted
	// TODO: add a remove function
	newNamedDataStore: function(modelName, isLegacyImporter) {
		if (this.modelName != null) {
			smLog("smDataStore.newNamedDataStore: error a named model can't create a new store.");
			return null; 
		}
		
		modelName = smUtils.toLowerCase(modelName);

		if (modelName!=null && (modelName == 'sxipper' || /\W/.test(modelName))) {
			smLog("smDataStore.newNamedDataStore: error bad name for model: " + modelName, L_ERROR);
			return null;
		}
		if (!modelName && !isLegacyImporter) {
			smLog("smDataStore.newNamedDataStore: error null model and not legacy importer!", L_ERROR);
			return null;
		}

		if (!this._storagePrefixSingletons[modelName]) {
			this._storagePrefixSingletons[modelName] = new smIDataStore(modelName, isLegacyImporter);
			smLog("smDataStore.newNamedDataStore: model " + modelName + " created.");
		} else {
			smLog("smDataStore.newNamedDataStore: model " + modelName + " already exists...");
		}

		return this._storagePrefixSingletons[modelName];
	},

	/* JSON
	 * 
	 *  JSON storage functions http://www.json.org/js.html
	 *  adopted from public domain code: http://www.json.org/json.js / 2007-10-10
	 *  updated to public domain code: http://www.json.org/json2.js / 2008-03-14
	 *  	the original source can be found /src/content/test/json
	 *  TODO: regularly check for updates to the original script and possible security problems
	 *  TODO: use FF3.0's native JSON
	 */
	JSON: {
		composeString: function (s) {
	        var m = {
	            '\b': '\\b',
	            '\t': '\\t',
	            '\n': '\\n',
	            '\f': '\\f',
	            '\r': '\\r',
	            '"' : '\\"',
	            '\\': '\\\\'
	        };
	        var r = /["\\\x00-\x1f\x7f-\x9f]/g;

			// If the string contains no control characters, no quote characters, and no
			// backslash characters, then we can simply slap some quotes around it.
			// Otherwise we must also replace the offending characters with safe
			// sequences.
			
			if (r.test(s)) {
			    return '"' + s.replace(r, function (a) {
			        var c = m[a];
			        if (c) {
			            return c;
			        }
			        c = a.charCodeAt();
			        return '\\u00' + Math.floor(c / 16).toString(16) +
			                                   (c % 16).toString(16);
			    }) + '"';
			}
			return '"' + s + '"';
			
		},
		
		composeNumber: function (n) {
			// JSON numbers must be finite. Encode non-finite numbers as null.
			
			return isFinite(n) ? String(n) : 'null';
		},
		
		composeBoolean: function (b) {
			return String(b);
		},
		
		composeArray: function (w) {
			var a = [],     // The array holding the partial texts.
			    i,          // Loop counter.
			    l = w.length,
			    v;          // The value to be stringified.
			
			// For each value in this array...
			
			for (i = 0; i < l; i += 1) {
			    v = this.compose(w[i]);
			    a.push(v);
			}
			
			// Join all of the member texts together and wrap them in brackets.
			
			return '[' + a.join(',') + ']';
		},

		composeObject: function (o) {
			var a = [],     // The array holding the partial texts.
			    i,          // Loop counter.
			    v;          // The value to be stringified.
			
			// For each key: value in this object...
			for (k in o) {
				if (typeof k === 'string' &&
						Object.prototype.hasOwnProperty.apply(o, [k])) {
					
					v = o[k];
					a.push(this.composeString(k) + ':' + this.compose(v));
					
				}
			}
			
			// Join all of the member texts together and wrap them in brackets.
			
			return '{' + a.join(',') + '}';
		},
		
		composeDate: function (d) {
			return '"' + String(d) + '"';
			/*			
			// Eventually, this method will be based on the date.toISOString method.
			
			function f(n) {
			// Format integers to have at least two digits.
				return n < 10 ? '0' + n : n;\
			}
			
			return '"' + d.getUTCFullYear()   + '-' +
					f(d.getUTCMonth() + 1) + '-' +
					f(d.getUTCDate())      + 'T' +
					f(d.getUTCHours())     + ':' +
					f(d.getUTCMinutes())   + ':' +
					f(d.getUTCSeconds())   + 'Z"';
			*/
		},

		compose: function (o) {
			if (typeof o == 'undefined')
				return 'undefined';
			else if (o === null)
				return 'null';
			//smLog('smDataStore.JSON.compose: typeof o: '+typeof o);
		    switch (typeof o) {
			    case 'object':
					// Serialize a JavaScript object value. Due to a specification error in ECMAScript,
					// typeof null is 'object', so watch out for that case.
					if (smIsArray(o)) {
		    			//smLog("smDataStore.JSON.compose: array");
						return this.composeArray(o);
					} else if (smIsDate(o)) {
		    			//smLog("smDataStore.JSON.compose: date");
						return this.composeDate(o);
					} else {
		    			//smLog("smDataStore.JSON.compose: object");
						return this.composeObject(o);
					}
			    	return 'null';
					break;

			    case 'string':
			    	return this.composeString(o);
			    case 'number':
			    	return this.composeNumber(o);
			    case 'boolean':
			    	return this.composeBoolean(o);
			    default:
			    	return 'null';
		    }

			return 'null';
		},

		/*
		 * parses string formated as JSON into a JavaScript object
		 */
		parse: function (text) {
			if (!text) {
				return null;
			}
			
            var j;

			// Parsing happens in three stages. In the first stage, we run the text against
			// a regular expression which looks for non-JSON characters. We are especially
			
			// concerned with '()' and 'new' because they can cause invocation, and '='
			// because it can cause mutation. But just to be safe, we will reject all
			// unexpected characters.
			
			// We split the first stage into 4 regexp operations in order to work around
			// crippling deficiencies in IE's and Safari's regexp engines. First we replace
			// all backslash pairs with '@' (a non-JSON character). Second, we replace all
			// simple value tokens with ']' characters. Third, we delete all open brackets
			// that follow a colon or comma or that begin the text. Finally, we look to see
			// that the remaining characters are only whitespace or ']' or ',' or ':' or '{'
			// or '}'. If that is so, then the text is safe for eval.
			if (/^[\],:{}\s]*$/.test(text.replace(/\\["\\\/bfnrtu]/g, '@').
                    replace(/"[^"\\\n\r]*"|true|false|null|undefined|-?\d+(?:\.\d*)?(:?[eE][+\-]?\d+)?/g, ']').
                    replace(/(?:^|:|,)(?:\s*\[)+/g, ''))) {
				
				// In the second stage we use the eval function to compile the text into a
				// JavaScript structure. The '{' operator is subject to a syntactic ambiguity
				// in JavaScript: it can begin a block or an object literal. We wrap the text
				// in parens to eliminate the ambiguity.
			 	try {
			 		// TODO: is about:blank an untrusted domain?
			 		// see http://mxr.mozilla.org/mozilla/source/js/src/xpconnect/loader/JSON.jsm
			 		var JSsandbox = new Components.utils.Sandbox('about:blank');
		            j = Components.utils.evalInSandbox('(' + text + ')', JSsandbox);
	                return j;
			 	}
			 	catch(e) {
		            smLog("smDataStore.JSON.parse: error " + e + " - '" + text + "'");
			 	}
            } else {
		            smLog("smDataStore.JSON.parse: error bad format - '" + text + "'");
            }
 
			// If the text is not JSON parseable, then a SyntaxError is thrown.
            //throw new SyntaxError('smDataStore.JSON.parse');
            
            // TODO: return empty array, object or null?
            return {};
        }
	}, // JSON finish

	/*
	 * FILE STORE
	 * 
	 *  File Storage functions
	 * 
	 * TODO: error check, santize, ...
	 * TODO: if write fails: warn user, schedule a later write, etc.
	 */	
	_filePrototype: {
		test: function() {
			return this.dataStore.modelName;
		},
		// constants
		FILE_FORMAT_JSON: 'json',

		FILE_TYPE_COMPLETE: 'complete',
		FILE_TYPE_APPEND: 'append',
		
		// returns the directory to store file in
		// storePath: is usually an array of folders that are inside sxipper directory
		//   but it can also be an array of a single nsIFile pointing to a directory
		//   anywhere on the filesystem
		_getFileDir: function(storePath) {
			if (this.dataStore.isLegacyImporter) {
				return smConfig.getLegacyLocalSxipperDirPath(storePath);
			}
			if (storePath.length==1 && (storePath[0] instanceof Components.interfaces.nsIFile)) {
				// TODO: QA on windows, QA if contains spaces
				return storePath[0].path;
			}
			// TODO: call assure only for write
			return smConfig.assureLocalSxipperDirExistence(storePath);
		},
		
		_jsonFile: function(storePath, filename) {
			// TODO: sanitize
			if (smConstants.invalidFilenameRegExp.test(filename)) {
				smLog("smDataStore.file._jsonFile: Invalid filename: [" +filename + "]");
				return null;
			}
			var fu = new smJslib.FileUtils(); // uses platform specific path delimiter
			var path = this._getFileDir(storePath);
			if (!fu.exists(path)) {
				return null;
			}
			if (this.dataStore.storagePrefix) {
				filename = this.dataStore.storagePrefix + '-' + filename;
			}
		    var filepath = fu.append(path, filename+'.json');
			var file = new smJslib.File(filepath);
			return file;
		},

		// generates the header to be included in all data store files
		_genFileHeader: function(fileType, isForceUnEncrypted) {
			var fileHeader = {
					id: smConstants.extensionId,
					version: smConstants.ledgerVersion,
					encrypted: isForceUnEncrypted ? false : smConstants.ledgerEncrypt,
					format: this.FILE_FORMAT_JSON,
					type: fileType
				};
			return fileHeader;
		},
		
		_createOutputStream: function(file) {
			//converterOutputStream -> bufferedOutputStream -> fileOutputStream
			var bos = Components.classes["@mozilla.org/network/buffered-output-stream;1"]
							.createInstance(Components.interfaces.nsIBufferedOutputStream);
			bos.init(file.mOutStream, 0x8000);
			
			var charset = "UTF-8";
			var cos = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
							.createInstance(Components.interfaces.nsIConverterOutputStream);
			cos.init(bos, charset, 0, 0x8000);
			return cos;
		},

		_writeSafely: function(file, string) {
			// TODO: make it buffered too
			
			var sos = Components.classes["@mozilla.org/network/safe-file-output-stream;1"]
							.createInstance(Components.interfaces.nsIFileOutputStream);
			
			// TODO: check if file permission are secure
			sos.init(file.nsIFile, 0x02 | 0x08 | 0x20, 0600, 0);
			sos.QueryInterface(Components.interfaces.nsISafeOutputStream);
			
			var charset = "UTF-8";
			var cos = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
							.createInstance(Components.interfaces.nsIConverterOutputStream);
			cos.init(sos, charset, 0, 0x8000);
			cos.writeString(string);

			cos.flush();
			sos.flush();
			sos.finish();
			file.close();
		},
			
		// delete file
		remove: function(storePath, filename) {
			var file = this._jsonFile(storePath, filename);
			return file.remove();
		},
		
		// serialize object and write it to a complete JSON file
		write: function (storePath, filename, object, isForceUnEncrypted) {
			var file = this._jsonFile(storePath, filename);

			var headerJSON = this.dataStore.JSON.compose(this._genFileHeader(this.FILE_TYPE_COMPLETE, isForceUnEncrypted));
			var dataString = headerJSON + "\n";

			var objJSON = this.dataStore.JSON.compose(object);
			if (!isForceUnEncrypted && smConstants.ledgerEncrypt) {
				// handle refusal of master password
				var encryptedValue = smUtils.encrypt(objJSON);
				if (!encryptedValue) {
					smLog("smDataStore.file.write: error encrypting data! master password missing?", L_ERROR);
		            throw "smDataStore.file.write: encryption failed.";
				}
				dataString += encryptedValue + "\n";
			} else {
				dataString += objJSON + "\n";
			}
			
			this._writeSafely(file, dataString);
		},
	

		// read a complete JSON file and return its content as an object
		readall: function (storePath, filename) {
			try {
				var file = this._jsonFile(storePath, filename);
			} catch (e) {
				//TODO
				smLog('dataStore.js:readall: Error in _jsonFile: '+e);
				return {};
			}
		    if (!file || !file.exists()) {
		    	return {};
		    }
		    
			//file.open('r');
			//var header = this.dataStore.JSON.parse(smUtils.readlineUTF8(file));
			//var content = smUtils.readlineUTF8(file);
			//file.close();

			// open an input stream from file
			var fis = Components.classes["@mozilla.org/network/file-input-stream;1"]
			                        .createInstance(Components.interfaces.nsIFileInputStream);
			fis.init(file.nsIFile, 0x01, 0400, 0);

			var charset = "UTF-8";
			const replacementChar = Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER;
			var istream = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
	                   .createInstance(Components.interfaces.nsIConverterInputStream);
	        istream.init(fis, charset, 1024, replacementChar);

			istream.QueryInterface(Components.interfaces.nsIUnicharLineInputStream);

			// read header line
			var line = {};
			var hasMore;
			hasMore = istream.readLine(line);
			var header = this.dataStore.JSON.parse(line.value);
			// read content line
			hasMore = istream.readLine(line);
			var content = line.value;

			// close file
		    istream.close();
			
			if (!header) {
				return null;
			}
			if (header.encrypted) {
				// TODO: what happens if no master password?
				return this.dataStore.JSON.parse(smUtils.decrypt(content));
			} else {
				return this.dataStore.JSON.parse(content);
			}
			return null;
		},

		// serialize object and rewrite it to a complete JSON file merging new data with the old file
		// TODO: error handling
		rewrite: function (storePath, filename, newData) {
			var oldData = this.readall(storePath, filename);
			var data = smUtils.mergeAssocArray(oldData, newData, true);
			this.write(storePath, filename, data);
			if (smConfig.getBoolPref('ledger.debug')) {
				smUtils.inspectObject({storePath: storePath, filename: filename, object: newData});
			}
		},

		// TODO: add version check 
		checkCompatibility: function(storePath, filename, type) {
			// TODO: write comptability checks for complete file as well
			if (type != this.FILE_TYPE_APPEND) {
				return true;
			}

			var file = this._jsonFile(storePath, filename);
			//file.open('r');
			//var header = this.dataStore.JSON.parse(smUtils.readlineUTF8(file));
			//file.close();

			// open an input stream from file
			var fis = Components.classes["@mozilla.org/network/file-input-stream;1"]
			                        .createInstance(Components.interfaces.nsIFileInputStream);
			fis.init(file.nsIFile, 0x01, 0400, 0);

			var charset = "UTF-8";
			const replacementChar = Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER;
			var istream = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
	                   .createInstance(Components.interfaces.nsIConverterInputStream);
	        istream.init(fis, charset, 1024, replacementChar);

			istream.QueryInterface(Components.interfaces.nsIUnicharLineInputStream);

			// read header line
			var line = {};
			var hasMore;
			hasMore = istream.readLine(line);
			var header = this.dataStore.JSON.parse(line.value);
			
			// close file
		    istream.close();

			if (header.id != smConstants.extensionId 
					|| header.version !=  smConstants.ledgerVersion
					|| header.format != this.FILE_FORMAT_JSON
					|| header.type != type) {
				return false;
			}
			if (header.encrypted == smConstants.ledgerEncrypt) {
				return true;		
			}

			// TODO: review the following code that upgrades encryption status
			smLog("smDataStore.file.checkCompatibility: TODO: encrypting/decrypting file...");
			var dataFile = this.iterator(storePath, filename);
			var item;
			while (item = dataFile.iterate()) {
				//smLog("smDataStore.file.checkCompatibility: "+smDataStore.JSON.compose(item));
				this.append(storePath, 'upgrading-'+filename, item);
			}
			dataFile.close();
			
			var upgradeFile = this._jsonFile(storePath, 'upgrading-'+filename);
			file.remove();
			upgradeFile.move(file);
			return true;
			
			//return false;
		},

		// detects type from the file itself
		// rewrites everything so that it can change encryption 
		moveToPath: function(storePath, filename, storePath2, isForceUnEncrypted) {
			// makes sure that storePath and storePath2 are different
			if (smUtils.isArraysEqual(storePath, storePath2)) {
				throw new Error("smDataStore.file.moveToPath: Error, storepath and storePath2 are equal: " + storePath.toSource());
				return;
			}
			var file = this._jsonFile(storePath, filename);
			//file.open('r');
			//var header = this.dataStore.JSON.parse(smUtils.readlineUTF8(file));
			//file.close();

			// open an input stream from file
			var fis = Components.classes["@mozilla.org/network/file-input-stream;1"]
			                        .createInstance(Components.interfaces.nsIFileInputStream);
			fis.init(file.nsIFile, 0x01, 0400, 0);

			var charset = "UTF-8";
			const replacementChar = Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER;
			var istream = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
	                   .createInstance(Components.interfaces.nsIConverterInputStream);
	        istream.init(fis, charset, 1024, replacementChar);

			istream.QueryInterface(Components.interfaces.nsIUnicharLineInputStream);

			// read header line
			var line = {};
			var hasMore;
			hasMore = istream.readLine(line);
			var header = this.dataStore.JSON.parse(line.value);
			
			// close file
		    istream.close();

			if (header.id != smConstants.extensionId 
					|| header.version !=  smConstants.ledgerVersion
					|| header.format != this.FILE_FORMAT_JSON
					|| (header.type != this.FILE_TYPE_COMPLETE && header.type != this.FILE_TYPE_APPEND)) {
				throw new Error("smDataStore.file.moveToPath: Error, bad file.");
				return;
			}

			smLog("smDataStore.file.moveToPath: moving " + filename + " from " + storePath + " to " + storePath2);
			if (header.type == this.FILE_TYPE_COMPLETE) {
				var data = this.readall(storePath, filename);
				this.write(storePath2, filename, data, isForceUnEncrypted);
			} else if (header.type == this.FILE_TYPE_APPEND) {
				// open source file
				var dataFile = this.iterator(storePath, filename);
				
				// delete the destination file if it exists
				var upgradeFile = this._jsonFile(storePath2, filename);
				upgradeFile.remove();
				
				var item;
				while (item = dataFile.iterate()) {
					this.append(storePath2, filename, item, isForceUnEncrypted);
				}
				dataFile.close();
			}
		},
		
		// serialize object and append it to a appendable JSON file
		// TODO: error handling
		// TODO: use safe write.
		append: function (storePath, filename, object, isForceUnEncrypted) {
			var file = this._jsonFile(storePath, filename);
			var isRequiringHeader = false;
		    if (!file.exists()) {
				file.create();
				file.open('w');
				isRequiringHeader = true;
		    } else {
				if (!isForceUnEncrypted && !this.checkCompatibility(storePath, filename, this.FILE_TYPE_APPEND)) {
					smLog("smDataStore.append: error compatibility check failed!");
					return false;
				}
				file.open('a');
		    }
			var os = this._createOutputStream(file);
			if (isRequiringHeader) {
				var headerJSON = this.dataStore.JSON.compose(this._genFileHeader(this.FILE_TYPE_APPEND, isForceUnEncrypted));
				os.writeString(headerJSON);
				os.writeString("\n");
			}

			if (smConfig.getBoolPref('ledger.debug')) {
				smUtils.inspectObject({storePath: storePath, filename: filename, object: object});
			}
			var objJSON = this.dataStore.JSON.compose(object);
		    if (!isForceUnEncrypted && smConstants.ledgerEncrypt) {
				//TODO better handling refusal of master password
				var encryptedValue = smUtils.encrypt(objJSON);
				if (!encryptedValue) {
					smLog("smDataStore.file.append: error encrypting data! master password missing?", L_ERROR);
				    os.flush();
				    os.close();
					file.close();
		            throw "smDataStore.file.append: encryption failed.";
				}
				os.writeString(encryptedValue);
				os.writeString("\n");
		    } else {
				os.writeString(objJSON);
				os.writeString("\n");
		    }
		    os.flush();
		    os.close();
			file.close();
		},
		
		// opens an appendable JSON file and returns an iterator function for the file contents
		//  methods of the the iterator function:
		//   .close: closes the file
		//   .iterate: returns the next JSON object and NULL if reached the end of file
		// TODO: make sure there's no limit on line length
		iterator: function (storePath, filename) {
			var file = this._jsonFile(storePath, filename);
			//file.open('r');
			// TODO: error handling
			// TODO: if bad header: backup, recreate...
			//var header = this.dataStore.JSON.parse(smUtils.readlineUTF8(file));
			
			if (!file.exists()) {
				smLog("smDataStore.file.iterator: no such file " + filename);
				return null;
			}

			// open an input stream from file
			var fis = Components.classes["@mozilla.org/network/file-input-stream;1"]
			                        .createInstance(Components.interfaces.nsIFileInputStream);
			fis.init(file.nsIFile, 0x01, 0400, 0);

			var charset = "UTF-8";
			const replacementChar = Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER;
			var istream = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
	                   .createInstance(Components.interfaces.nsIConverterInputStream);
	        istream.init(fis, charset, 1024, replacementChar);

			istream.QueryInterface(Components.interfaces.nsIUnicharLineInputStream);

			// read header line
			var line = {};
			var hasMore;
			hasMore = istream.readLine(line);
			var header = this.dataStore.JSON.parse(line.value);
			
			var funcHeader = function(header) {
				return function() {
					return header;
				};
			}
			/*
			var funcEOF = function(file) {
				return function() {
					return file.EOF;
				};
			}
			*/
			var funcIterate = function(istream, json, encrypted) {
				return function() {
					//var content = smUtils.readlineUTF8(file);
					
					// read content line
					hasMore = istream.readLine(line);
					var content = line.value;
					
					// return null if the file is finished, even if incomplete file
					if (!hasMore && !content) {
						return null;
					}
					if (!content) {
						smLog("smDataStore.file.iterator: error emtpy line");
						return {};
					}
					if (encrypted) {
						return json.parse(smUtils.decrypt(content));
					} else {
						return json.parse(content);
					}
					return null;
				};
			}
			var funcClose = function(istream) {
				return function() {
					//return file.close();
					
					// close file
				    istream.close();
				}
			}
			return {
				getHeader: funcHeader(header),
				//EOF: funcEOF(file),
				iterate: funcIterate(istream, this.dataStore.JSON, header.encrypted),
				close: funcClose(istream) 
			}
		}
		/*
		,
		
		test: function() {
			var a={ 
				test: 1, 
				checking :"string here! \' \" \n " , 
				aBoolean: true, 
				"anArray": ["1", 2, 3, 4+5], 
				"null": null,
				date: new Date(),
				unicode: '\u0646'
			};
			this.append('test', a);
	
			var i=this.iterator('test');
			var z;
			while (z=i.iterate()) {
				smLog("smDataStore.file.test: "+this.dataStore.JSON.compose(z));
			}
			i.close();
			
			this.rewrite('datastore', {lastrun: new Date()});
			var sxst = this.readall('datastore');
			smLog("smDataStore.file.test: lastrun: " + sxst.lastrun);
		}
		*/
	},
	
	/*
	 * LEDGER
	 * 
	 * 	Ledger captures all the events happening and store them based on type and domain they happend on.
	 * 	External event processors can be attached to ledger to consume events and possbily put them back 
	 *  as processed data into ledger.
	 */
	 // TODO: add a functionality 
	_ledgerPrototype: {
		// TODO: define as const?
		LEDGER_TYPE: {
			ALL_DOMAIN_LIST: 'allDomainList',
			DOMAIN_EVENT: 'domainEvent',
			DOMAIN_STATE: 'domainState'
		},
		
		// TODO: add login successful/unsuccessful. same for registration (form_fill)
		DOMAIN_EVENT_TYPE: {
			FORM_FILL: 'form_fill',
		 	LOGIN: 'login',
		 	LOGIN_UNSUCCESSFUL: 'login_unsuccessful',
			FORM_DISCOVERY: 'form_discovery'
		},
				
		DOMAIN_STATE_TYPE: {
			LAST_LOGIN: 'last_login',
			LOGIN_DATA: 'login_data',
			LEDGER: 'ledger'
		},
		
		_genEventId: function() {
			return smUtils.generateUuid();
		},

		domainEvent: function(domain, type, data) {
			var o = {
				id: this._genEventId(),
				date: new Date(),
				type: type,
				data: data
			};

			if (smConfig.getBoolPref('ledger.enabled')) {
				// TODO: filter out data['fields'] with types that should not be stored (e.g CC when disabled)
				
				//append event to domain ledger file
				this.dataStore.file.append(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_EVENT + '-' + domain, o);
	
				//mark domain name in list of all domains
				var domainObject={};
				domainObject[domain]=true;
				this.dataStore.file.rewrite(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.ALL_DOMAIN_LIST, domainObject);
			}
			
			if (this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT] && this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT][type]) {
				for (var i=0; i<this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT][type].length; i++) {
					var processorObject = this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT][type][i];
					processorObject.ledgerDomainEvent.call(processorObject, type, domain, o);
				}
			}
			
		},
		
		setDomainState: function(domain, type, data) {
//			if (!smConfig.getBoolPref('ledger.enabled')) {
//				return;
//			}
			// add data to domain state ledger file
			var o = {};
			o[type] = data;
			this.dataStore.file.rewrite(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_STATE + '-' + domain, o);

			//mark domain name in list of all domains
			var domainObject={};
			domainObject[domain]=true;
			this.dataStore.file.rewrite(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.ALL_DOMAIN_LIST, domainObject);

//			TODO: implement processor for domain state events
//			if (this.processorRegistry[this.LEDGER_TYPE.DOMAIN_STATE] && this.processorRegistry[this.LEDGER_TYPE.DOMAIN_STATE][type]) {
//				for (var i=0; i<this.processorRegistry[this.LEDGER_TYPE.DOMAIN_STATE][type].length; i++) {
//					this.processorRegistry[this.LEDGER_TYPE.DOMAIN_STATE][type][i].call(this, domain, o);
//				}
//			}

		},

		getDomainState: function(domain, type) {
			var state=this.dataStore.file.readall(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_STATE + '-' + domain);
			if (typeof state[type] == 'undefined') {
				return null;
			}
			return state[type];
		},

		// Processors are functions that consume an event of a certain type
		// They are registered in this.processorRegistry
		processorRegistry: {},
				
		init: function() {
			this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT] = {};
			this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT][this.DOMAIN_EVENT_TYPE.LOGIN] = [
					this.dataStore.passwordManager
				];
			this.processorRegistry[this.LEDGER_TYPE.DOMAIN_EVENT][this.DOMAIN_EVENT_TYPE.LOGIN_UNSUCCESSFUL] = [
					this.dataStore.passwordManager
				];

			this.processorRegistry[this.LEDGER_TYPE.DOMAIN_STATE] = {};
		},
		
		deleteAll: function() {
			var du = new smJslib.DirUtils();
			var fu = new smJslib.FileUtils();
			var ledgerPath = smConfig.assureLocalSxipperDirExistence(smConstants.ledgerLocalPathArray);
			smUtils.removeDirectory(ledgerPath);
		},

		dumpDomainEvents: function(domain) {
			var i=this.dataStore.file.iterator(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_EVENT + '-' + domain);
			var event;
			while (event = i.iterate()) {
				smLog("smDataStore.ledger.dumpDomainEvents ["+domain+"]: "+this.dataStore.JSON.compose(event));
			}
			i.close();
		},

		dumpDomainState: function(domain) {
			var state=this.dataStore.file.readall(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_STATE + '-' + domain);
			smLog("smDataStore.ledger.dumpDomainState ["+domain+"]: "+this.dataStore.JSON.compose(state));
		},
		
		inspectDomainEvents: function() {
			var domains=this.dataStore.file.readall(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.ALL_DOMAIN_LIST);
		
			smUtils.inspectObject(domains);
			for (var domain in domains) {
				var events = [];

				domains[domain]=events;

				var i=this.dataStore.file.iterator(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_EVENT + '-' + domain);
				if (!i) {
					continue;
				}
				var event;
				while (event = i.iterate()) {
					events.push(event);
				}
				i.close();
			}			

		},

		inspectDomainState: function(domain) {
			var domains=this.dataStore.file.readall(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.ALL_DOMAIN_LIST);
		
			smUtils.inspectObject(domains);
			for (var domain in domains) {
				var state=this.dataStore.file.readall(smConstants.ledgerLocalPathArray, this.LEDGER_TYPE.DOMAIN_STATE + '-' + domain);
				domains[domain] = state;
			}
		}
		/*
		, 
		
		test: function() {
			
			this.domainEvent('pooyak.com', 
						this.DOMAIN_EVENT_TYPE.FORM_FILL, 
						{hello: "a", check:123}
					);
			this.dumpDomainEvents('pooyak.com');
			
			
			this.dumpDomainState('pooyak.com');
		}
		*/
		
				
	},

	_personalPrototype: {
		TYPE: {
			PERSONA: 'persona',
			PROPERTY: 'property',
			MULTIFIELD: 'multifield'
		},
		
		init: function() {
			var personaPrototype = function() { };
			personaPrototype.prototype = this._personaPrototype;
			this.persona = new personaPrototype();

			var propertyPrototype = function() { };
			propertyPrototype.prototype = this._propertyPrototype;
			this.property = new propertyPrototype();

			var multiFieldPrototype = function() { };
			multiFieldPrototype.prototype = this._multiFieldPrototype;
			this.multiField = new multiFieldPrototype();

			this.persona._delayedRunner = new smIDelayedRunner(smConstants.delayedDataStoreWriteTime);
			this.property._delayedRunner = new smIDelayedRunner(smConstants.delayedDataStoreWriteTime);
			this.multiField._delayedRunner = new smIDelayedRunner(smConstants.delayedDataStoreWriteTime);

			// allow inner objects to access datastore
			this.persona.dataStore = this.dataStore;
			this.property.dataStore = this.dataStore;
			this.multiField.dataStore = this.dataStore;
			
		},
	
		flushAll: function() {
			smLog("smDataStore.personal.flushAll: writing all the caches.");
			this.persona.flush();
			this.property.flush();
			this.multiField.flush();
		},
		
		// clear caches and cancel delayed writes
		clearAllCache: function() {
			smLog("smDataStore.personal.clearCache: clearing all the caches.");
			this.persona.clearCache();
			this.property.clearCache();
			this.multiField.clearCache();
		},
		
		// TODO: add case insensitive values to persona and properties and their actions: lookup, delete
		// TODO: separate functions that access _cache and those that don't
		
		_personaPrototype: {
			_genId: function() {
				return smUtils.generateUuid();
			},
			
			_cache: null,
//			 Persona data cache structure:
//				 { "#id': {
//						label: 'Persona Label',
//						created: DATE,
//						icon: 'http://ICON/URL',
//			 			data: { 'propertyType1': {value: 'value', created: DATE},
// 								'propertyType2': {value: 'value', created: DATE},
//								...
//							}
//			 		},
//			 	 ...
//				 }

			_cacheDirty: false,

			_checkCache: function() {
				if (!this._cache) {
					this._cache = this.dataStore.file.readall(smConstants.personalDataLocalPathArray, this.dataStore.personal.TYPE.PERSONA);
				}
				// TODO: if null again create an empty file
			},
			
			_delayedRunner: null,

			_writeCache: function(isQuit) {
				this._checkCache();

				var self = this;
				var writerFunc = function() {
						smLog("smDataStore.personal.persona._writeCache: writing...");
						self.dataStore.file.write(smConstants.personalDataLocalPathArray, 
											self.dataStore.personal.TYPE.PERSONA, 
											self._cache);
						//smUtils.inspectObject(this._cache);
						// if here write was successful: write throws an error if it fails 
						this._cacheDirty = false;
					};
					
				if (smConstants.delayedDataStoreWriteEnabled && !isQuit) {
					this._cacheDirty = true;
					this._delayedRunner.run(writerFunc);
				} else {
					writerFunc();
				}
			},
			
			flush: function() {
				if (!this._cacheDirty) {
					return;
				}
				this._writeCache(true);
			},
			
			clearCache: function() {
				if (smConstants.delayedDataStoreWriteEnabled) {
					this._delayedRunner.cancel();
				}
				this._cache = null;
				this._cacheDirty = false;
			},
			
			create: function(label) {
				this._checkCache();

				var uniqueId = this._genId();
				this._cache[uniqueId] = {
								'label': label, 
								'created': new Date(), 
								'data': {}
							};
				this._writeCache();

				return uniqueId;
			},
			
			remove: function(personaId) {
				this._checkCache();
				
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.remove: error no such personaid: " + personaId);
					return;
				}
				
				delete this._cache[personaId];
				
				this._writeCache();
			},
			
			getPersonaIds: function() {
				this._checkCache();
				
				var personaIds = [];
				for (var personaId in this._cache) {
					personaIds.push(personaId)
				}
				
				return personaIds;
			},
			
			getLabel: function(personaId) {
				this._checkCache();
				
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.getLabel: error no such personaid: " + personaId);
					return null;
				}
				
				return this._cache[personaId].label;
			},
			
			setLabel: function(personaId, label) {
				this._checkCache();
				
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.setLabel: error no such personaid: " + personaId);
					return;
				}
				
				this._cache[personaId].label = label;

				this._writeCache();
			},

			getIcon: function(personaId) {
				this._checkCache();
				
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.getIcon: error no such personaid: " + personaId);
					return null;
				}
				
				return this._cache[personaId].icon;
			},
			
			setIcon: function(personaId, icon) {
				this._checkCache();
				
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.setIcon: error no such personaid: " + personaId);
					return;
				}
				
				// TODO: validate
            	//if (icon.indexOf(smConfig.PERSONA_ICON_URI_PREFIX) != 0) {
            	//}
				
				this._cache[personaId].icon = icon;

				this._writeCache();
			},

			deleteAll: function() {
				// TODO: delete all the properties as well
				this._cache = {};
				this._writeCache();
			},
			
			getPropertyValue: function(personaId, propertyType) {
				this._checkCache();
					
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.getPropertyValue: error no such personaid: " + personaId);
					return null;
				}

				if (typeof this._cache[personaId].data[propertyType]=='undefined') {
					return null;
				}
				
				// TODO: is string only fine?
				return String(this._cache[personaId].data[propertyType].value);
			},
			
			// TODO: sanitize, ...
			setPropertyValue: function(personaId, propertyType, value) {
				this._checkCache();
					
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.setPropertyValue: error no such personaid: " + personaId);
					return;
				}

				if (!this.dataStore.personal.property.isValid(propertyType, value)) {
					smLog("smDataStore.personal.persona.setPropertyValue: invalid property for personaid: " + personaId);
					return;
				}
				
				var currentValue = this.getPropertyValue(personaId, propertyType);
				if (currentValue == null || currentValue != value) {
					// TODO: overwrite if already exists?
					//if (this._cache[personaId].data[propertyType]) {
					//	this._cache[personaId].data[propertyType] = {};
					//}
					
					// TODO: is it fine that we only store strings?
					this._cache[personaId].data[propertyType]= {value: String(value), created: new Date()};
					this._writeCache();
				}
				
				// also add it to properties if not empty or null
				// empty and null makes sense for a persona property but not as a free one
				if (value!=null && value!="") {
					this.dataStore.personal.property.addItem(propertyType, value);
				}
			},

			getAllProperties: function(personaId) {
				this._checkCache();
					
				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.getAllProperties: error no such personaid: " + personaId);
					return;
				}
				
				var properties = {};
				for (var propertyType in this._cache[personaId].data) {
					properties[propertyType] = this._cache[personaId].data[propertyType].value;
				}
				return properties;
			},

			deleteProperty: function(personaId, propertyType) {
				this._checkCache();

				if (!this._cache[personaId]) {
					smLog("smDataStore.personal.persona.deleteProperty: error no such personaid: " + personaId);
					return;
				}

				if (!this._cache[personaId].data[propertyType] ) {
					smLog("smDataStore.personal.property.remove: error no such property " + propertyType + " in " + personaId);
					return;
				}

				delete this._cache[personaId].data[propertyType];

				this._writeCache();
			},

			deletePropertyValueFromAllPersonas: function(propertyType, value) {
				//this._checkCache();

				var personaIds = this.getPersonaIds();
				for (var i=0; i<personaIds.length; i++) {
					var personaPropertyValue = this.getPropertyValue(personaIds[i], propertyType);
					if (personaPropertyValue!=null && personaPropertyValue == value) {
						this.deleteProperty(personaIds[i], propertyType);
					}
				}

				//this._writeCache();
			},
			
			// generates a multi field entry of type compoundPropertyType from this persona 
			convertToMultifield: function(personaId, compoundPropertyType) {
//				if (compoundPropertyType == smConstants.sxipperCreditCardCompound) {
//					if (!smConfig.getBoolPref("isCreditcardStorageEnabled")) {
//						smLog("smDataStore.personal.persona.convertToMultifield: storing CC information is disabled");
//						return null;
//					}
//				}

				var props = smGroupSchema.getGroup(compoundPropertyType).contents();
		
				var entry = {count: 0, items: {}};
		
				for (var k=0; k < props.length; k++) {
					var personalValue = this.getPropertyValue(personaId, props[k]);
					
					if (smUtils.isEmptyString(personalValue)) {
						continue;
					}
					
			 		if (!smGroupSchema.isValidSelectionPropertyValue(props[k], personalValue)) {
			 			//smLog("smDataStore.personal.persona.convertToMultifield: error bad selection item in persona prop=value : "
			 			//		+ props[k] + "=" + personalValue);
			 			continue;
			 		}

					entry.count++;
					entry.items[props[k]] = personalValue; 
				}
				//smLog("smDataStore.personal.persona.convertToMultifield: type:" + compoundPropertyType + " items:"+entry.count);
				if (entry.count < smConstants.minNumberOfMultiFields) {
					return null;
				}
				// TODO: do we need this now that multiField itself has validation?
				if (compoundPropertyType == smConstants.sxipperCreditCardCompound) {
					entry.items[smConstants.sxipperCCLabelLocal] = this.getLabel(personaId);
				}
				
				return entry.items;
			},
			
			deleteExactMultifieldFromAllPersonas: function(compoundPropertyType, items) {
				//this._checkCache();

				var personaIds = this.getPersonaIds();
				for (var i=0; i<personaIds.length; i++) {
					var personaMultiField = this.convertToMultifield(personaIds[i], compoundPropertyType);
					if (personaMultiField!=null 
							&& this.dataStore.personal.multiField.isDuplicateItems(personaMultiField, items, false, true)) {
						smLog("smDataStore.personal.persona.deleteExactMultifieldFromAllPersonas: deleting " +
									" compoundPropertyType: " + compoundPropertyType +
									" from persona: " + this.getLabel(personaIds[i])
								);
						for (var propertyType in items) {
							this.deleteProperty(personaIds[i], propertyType);
						}
					}
				}
				//this._writeCache();
			},
			
			hasPersonasMultiValue: function(compoundPropertyType, items) {
				var personaIds = this.getPersonaIds();
				for (var i=0; i<personaIds.length; i++) {
					var personaMultiField = this.convertToMultifield(personaIds[i], compoundPropertyType);
					if (personaMultiField!=null 
							&& this.dataStore.personal.multiField.isDuplicateItems(personaMultiField, items, false, true)) {
						smLog("smDataStore.personal.persona.hasPersonaMultiValue:  " +
								" compoundPropertyType: " + compoundPropertyType +
								" from persona: " + this.getLabel(personaIds[i])
						);
						return true;
					}
				}
				return false;
			},
			
			inspect: function () {
				//this._cache = null;
				this._checkCache();
				smUtils.inspectObject({
					data: this._cache,
					json: this.dataStore.JSON.compose(this._cache)
				});
			} 
		},
		
		_propertyPrototype: {
			_genId: function() {
				return smUtils.generateUuid();
			},
			
			_cache: null,
//			 Property data cache structure:
//
//				{ 
// 					'propertyType1': {
// 									'value1': {created: DATE}, 
// 									'value2': {created: DATE}, 
// 									... },
//					'propertyType2': {'value1': {created: DATE}, 'value2': {created: DATE}, ... }, 
//					...
//				}

			_cacheDirty: false,

			_checkCache: function() {
				if (!this._cache) {
					this._cache = this.dataStore.file.readall(smConstants.personalDataLocalPathArray, this.dataStore.personal.TYPE.PROPERTY);
				}
			},

			_delayedRunner: null,

			_writeCache: function(isQuit) {
				this._checkCache();

				var self = this;
				var writerFunc = function() {
						smLog("smDataStore.personal.property._writeCache: writing...");
						self.dataStore.file.write(smConstants.personalDataLocalPathArray, 
											self.dataStore.personal.TYPE.PROPERTY, 
											self._cache);
						//smUtils.inspectObject(this._cache);
						this._cacheDirty = false;
					};
					
				if (smConstants.delayedDataStoreWriteEnabled && !isQuit) {
					this._cacheDirty = true;
					this._delayedRunner.run(writerFunc);
				} else {
					writerFunc();
				}
			},
			
			flush: function() {
				if (!this._cacheDirty) {
					return;
				}
				this._writeCache(true);
			},
			
			clearCache: function() {
				if (smConstants.delayedDataStoreWriteEnabled) {
					this._delayedRunner.cancel();
				}
				this._cache = null;
				this._cacheDirty = false;
			},
			
			// TODO: separate isValidToWrite and isValidToRead
			// TODO: isValidRead: fragments, non-confirmed openids, disposable emails, etc.
			// TODO: isValidWrite: emails that are not valid, start with dispoable->, ...
			isValid: function(propertyType, value) {
			    if (propertyType.indexOf('#') != -1) {
		 			smLog("smDataStore.personal.property.isValid: error fragments not allowed: "+propertyType);
		 			return false;
			    }
		 		if (smGroupSchema.isReadonlyProperty(propertyType)) {
		 			smLog("smDataStore.personal.property.isValid: error not valid. readonly property!: "+propertyType);
		 			return false;
		 		}
		 		if (!smGroupSchema.isValidSelectionPropertyValue(propertyType, value)) {
		 			smLog("smDataStore.personal.property.isValid: error not valid. selection property=value: "+propertyType+'='+value);
		 			return false;
		 		}
		 		// make sure the file exists
		 		if (propertyType == smConstants.propertyMediaImage) {
		            var imageInfo = smUtils.getPersonalImageInfo(value);
					if (!imageInfo) {
		 				smLog("smDataStore.personal.property.isValid: error not valid. bad image : "+propertyType+'='+value);
						return false;
					}
		 		}

		 		if (propertyType == smConstants.sxipperOpenID 
		 		 && (smConstants.isSxipperPublicOpenIdRegExp.test(value)
		 		  || smConstants.directedIdRegexp.test(value))) {
	 				smLog("smDataStore.personal.property.isValid: no public openid : "+propertyType+'='+value);
					return false;
		 		}
		 		
		 		// deprecated
		 		if (value == smConstants.nymUnverifiedString) {
		 			return false;
		 		}
		 		
		 		return true;
			},

			// TODO: sanitize, ...
			addItem: function(propertyType, value) {
				this._checkCache();
				
				if (!this.isValid(propertyType, value)) {
					return;
				}
				if (this.isItemExists(propertyType, value)) {
					//smLog("smDataStore.personal.property.create: propertyType=value already exists: " + propertyType + '=' + value);
					//this.deleteItem(propertyType, value);
					return;
				}

				if (!this._cache[propertyType]) {
					this._cache[propertyType] = {};
				}

				this._cache[propertyType][value] = {created: new Date()};
				this._writeCache();
			},
			
			deleteItem: function(propertyType, value) {
				this._checkCache();

				if (!this._cache[propertyType] || !this._cache[propertyType][value]) {
					smLog("smDataStore.personal.property.remove: error no such propertyType=value " + propertyType + "=" + value);
					return;
				}

				delete this._cache[propertyType][value];

				var remainingCount = smUtils.lengthAssociativeArray(this._cache[propertyType]);
				if (remainingCount==0) {
					delete this._cache[propertyType];
				}
				
				this._writeCache();

				//  changed: always delete it from personas even if in multifields
				//if (!this.dataStore.personal.multiField.isPropertyExists(propertyType, value)) {
					this.dataStore.personal.persona.deletePropertyValueFromAllPersonas(propertyType, value);
				//}
			},

			deleteAll: function() {
				// TODO: delete individual items as well
				this._cache = {};
				this._writeCache();
			},

			isItemExists: function(propertyType, value) {
				this._checkCache();
				
				if (!this._cache[propertyType]) {
					return null;
				}
				if (!this._cache[propertyType][value]) {
					return null;
				}
				
				return typeof this._cache[propertyType][value] != 'undefined';
			},
			
			getPropertyItems: function(propertyType) {
				this._checkCache();
				
				if (!this._cache[propertyType]) {
					return [];
				}

				var result = [];
				for (var value in this._cache[propertyType]) {
					result.push({propertyType: propertyType, value: value});
				}
				
				return result;
			},
			
			// propertyType can be null which returns all the items
			// TODO: case insensitive
			getItems: function(propertyType) {
				if (propertyType) {
					return this.getPropertyItems(propertyType);
				} 

				var result = []
				this._checkCache();

				for (var propertyType in this._cache) {
					result = result.concat(this.getPropertyItems(propertyType));
				}
				
				return result;
			},
		
			inspect: function () {
				//this._cache = null;
				this._checkCache();
				smUtils.inspectObject({
					data: this._cache,
					json: this.dataStore.JSON.compose(this._cache)
				});
			} 
		},
		
		// TODO: create a model.js wrapper for multifield (see smIdoGroup) (?)
		_multiFieldPrototype: {
			_genId: function() {
				return smUtils.generateUuid();
			},
			
			_cache: null, 
//			 Multi field data cache structure:
//			 { 'http://./compound/address': { 
//			 			'#id': { 'propertyType1': 'value1', 'propertyType2': 'value2', ...}, 
//			 			...
//			 		},
//			 	 ...
//			 }

			_cacheDirty: false,
			 
			_checkCache: function() {
				if (!this._cache) {
					this._cache = this.dataStore.file.readall(smConstants.personalDataLocalPathArray, this.dataStore.personal.TYPE.MULTIFIELD);
				}
			},
			
			_delayedRunner: null,

			_writeCache: function(isQuit) {
				this._checkCache();

				var self = this;
				var writerFunc = function() {
						smLog("smDataStore.personal.multiField._writeCache: writing...");
						self.dataStore.file.write(smConstants.personalDataLocalPathArray, 
											self.dataStore.personal.TYPE.MULTIFIELD, 
											self._cache);
						//smUtils.inspectObject(this._cache);
						this._cacheDirty = false;
					};
					
				if (smConstants.delayedDataStoreWriteEnabled && !isQuit) {
					this._cacheDirty = true;
					this._delayedRunner.run(writerFunc);
				} else {
					writerFunc();
				}
			},

			flush: function() {
				if (!this._cacheDirty) {
					return;
				}
				this._writeCache(true);
			},
			
			clearCache: function() {
				if (smConstants.delayedDataStoreWriteEnabled) {
					this._delayedRunner.cancel();
				}
				this._cache = null;
				this._cacheDirty = false;
			},
			
			getAll: function() {
				this._checkCache();
				
				return this._cache;
			},

			getCompound: function(propertyType) {
				this._checkCache();
				
				if (this._cache[propertyType]) {
					return this._cache[propertyType];
				}
				
				return null;
			},
			
			// TODO: add partial validation: keeping the rest of valid data
			isValid: function(compoundPropertyType, items) {
//				if (compoundPropertyType == smConstants.sxipperCreditCardCompound) {
//					if (!smConfig.getBoolPref("isCreditcardStorageEnabled")) {
//						smLog("smDataStore.personal.multiField.isValid: storing CC information is disabled");
//						return false;
//					}
//				}

				var count = 0;
				for (var propertyType in items) {
					// check if compound and propertyType are valid and match each other
					if (!smGroupSchema.isValidMultiField(compoundPropertyType, propertyType)) {
						smLog("smDataStore.personal.multiField.isValid: error invalid type: " + compoundPropertyType + ':' + propertyType);
						return false;
					}

					// don't allow invalid selectlists
			 		if (!smGroupSchema.isValidSelectionPropertyValue(propertyType, items[propertyType])) {
						smLog("smDataStore.personal.multiField.isValid: error invalid value: " + propertyType + '=' + items[propertyType]);
			 			return false;
			 		}

					count++;
				}
				if (count > 0) {
					return true;
				}
				smLog("smDataStore.personal.multiField.isValid: error no entry: " + compoundPropertyType);
				return false;
			},
			
			// per default we allow items2 to be a subset of items1
			isDuplicateItems: function(items1, items2, isCaseInsensetive, isSubsetDisabled) {
				var totalItems2 = 0;
				for (var propertyType in items2) {
					if (propertyType == smConstants.sxipperCCLabelLocal) {
						continue;
					}
					totalItems2++;
				}

				var totalItems1 = 0;
				var matchedPropsCount = 0;
				var matchedValuesCount = 0;
				
				for (var propertyType in items1) {
					if (propertyType == smConstants.sxipperCCLabelLocal) {
						continue;
					}

					totalItems1++;

					if (items2[propertyType]) {
						matchedPropsCount++;
						
						if (isCaseInsensetive) {
							if (smUtils.toLowerCase(items2[propertyType]) == 
									smUtils.toLowerCase(items1[propertyType])) {
								matchedValuesCount++;
							}
						} else {
							if (items2[propertyType] == items1[propertyType]) {
								matchedValuesCount++;
							}
						}
					}
				}
				
				// if both entries have the same number of items
				//  and all these items match the same name and value then they are equal
				if ((!isSubsetDisabled && totalItems2 <= totalItems1 || totalItems2 == totalItems1) && // subset
						totalItems2 == matchedPropsCount &&
						matchedPropsCount == matchedValuesCount) {
					return true;
				}
				return false;
				
			},
			
			isDuplicate: function(compoundPropertyType, items) {
				this._checkCache();
				
				if (!this._cache[compoundPropertyType]) {
					return false;
				}

				for (var id in this._cache[compoundPropertyType]) {
					if (this.isDuplicateItems(this._cache[compoundPropertyType][id], items)) {
						return true;
					}
				}

				return false;
			},
		
			addItem: function(compoundPropertyType, items) {
				this._checkCache();
				
				//smLog("smDataStore.personal.multiField.addItem: items: "+items.toSource());
				if (!this.isValid(compoundPropertyType, items)) {
					smLog("smDataStore.personal.multiField.addItem: invalid item");
					return null;
				} else if (this.isDuplicate(compoundPropertyType, items)) {
					smLog("smDataStore.personal.multiField.addItem: item already exists");
					return null;
				}
				
				if (!this._cache[compoundPropertyType]) {
					this._cache[compoundPropertyType] = {};
				}
				var uniqueId = this._genId();
				this._cache[compoundPropertyType][uniqueId] = items;
				this._writeCache();

				// also add it to properties
				for (var propertyType in items) {
					this.dataStore.personal.property.addItem(propertyType, items[propertyType]);
				}
				
				return uniqueId;
			},
			
			deleteCompound: function(compoundPropertyType) {
				// TODO: delete actual items
				this._checkCache();
				if (!this._cache[compoundPropertyType]) {
					smLog("smDataStore.personal.multiField.deleteItem: error no such compundType: " + compoundPropertyType);
					return;
				}

				delete this._cache[compoundPropertyType];

				this._writeCache();
			},

			deleteItem: function(compoundPropertyType, id) {
				this._checkCache();
				if (!this._cache[compoundPropertyType] || !this._cache[compoundPropertyType][id]) {
					smLog("smDataStore.personal.multiField.deleteItem: error no such propertyType:id " + compoundPropertyType+":" + id);
					return;
				}
				
				// first delete this in personas if it matches exactly too
				this.dataStore.personal.persona.deleteExactMultifieldFromAllPersonas(compoundPropertyType, this._cache[compoundPropertyType][id]);				
				
				// finally delete it
				delete this._cache[compoundPropertyType][id];

				var remainingPropsCount = smUtils.lengthAssociativeArray(this._cache[compoundPropertyType]);
				if (remainingPropsCount==0) {
					delete this._cache[compoundPropertyType];
				}
				this._writeCache();
			},

			isPropertyExists: function(propertyType, value) {
				this._checkCache();
				
				var compoundPropertyType = smGroupSchema.getMultiFieldParentCompound(propertyType);
				if (!compoundPropertyType) {
					return false;
				}
				
				if (!this._cache[compoundPropertyType]) {
					return false;
				}

				for (var id in this._cache[compoundPropertyType]) {
					//if (this.isDuplicateItems(this._cache[compoundPropertyType][id], items)) {
					if (typeof this._cache[compoundPropertyType][id][propertyType] != 'undefined'
						&& this._cache[compoundPropertyType][id][propertyType] == value) {
						return true;
					}
				}

				return false;
			},
		
			deleteAll: function() {
				this._cache = {};
				this._writeCache();
			},
			
			hasPersonasMultiValue: function(compoundPropertyType, id) {
				this._checkCache();
				if (!this._cache[compoundPropertyType] || !this._cache[compoundPropertyType][id]) {
					smLog("smDataStore.personal.multiField.deleteItem: error no such propertyType:id " + compoundPropertyType+":" + id);
					return null;
				}
				var items = this._cache[compoundPropertyType][id];
				return this.dataStore.personal.persona.hasPersonasMultiValue(compoundPropertyType, items);
			},
			
			inspect: function () {
				//this._cache = null;
				this._checkCache();
				smUtils.inspectObject({
					data: this._cache,
					json: this.dataStore.JSON.compose(this._cache)
				});
			}
		
		}
		
	},
	
	// TODO: sync with Firefox password manager
	_passwordManagerPrototype: {
		init: function() {
			
		},
		
		_readDomainLoginData: function(domain) {
			var loginData = this.dataStore.ledger.getDomainState(domain,
						this.dataStore.ledger.DOMAIN_STATE_TYPE.LOGIN_DATA
					);
			if (!loginData) {
				loginData = [];
			}
			
			return loginData;
		},
		
		_writeDomainLoginData: function(domain, loginData) {
			this.dataStore.ledger.setDomainState(domain,
				this.dataStore.ledger.DOMAIN_STATE_TYPE.LOGIN_DATA, 
				loginData
			);
		},

		_userEntryMatches: function(userEntry, hostname, users, passwords, formActionHostURL, httpRealm, isUserPassOnlyMatch) {
			if ((isUserPassOnlyMatch ||
						(userEntry.hostname == hostname &&
						userEntry.formActionHostURL == formActionHostURL &&
						userEntry.httpRealm == httpRealm)) &&
					smUtils.isArraysEqual(userEntry.users, users) &&
					smUtils.isArraysEqual(userEntry.passwords, passwords)) {
				return true;
			}
			return false;
		},
		
		_findUserInLoginData: function(loginData, hostname, users, passwords, formActionHostURL, httpRealm) {
			for (var i=0; i<loginData.length; i++) {
				if (this._userEntryMatches(loginData[i], hostname, users, passwords, formActionHostURL, httpRealm)) {
					return loginData[i];
				}
			}
			return null;
		},
		
		findUser: function(hostname, users, passwords, formActionHostURL, httpRealm) {
			if (!smIsArray(users)) {
				users = [users];
			}
			if (!smIsArray(passwords)) {
				passwords = [passwords];
			}
			var userDomain = smUtils.getRootDomain(hostname);
			var loginData = this._readDomainLoginData(userDomain);
			
			return this._findUserInLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);
		},

		_findAllUsersInLoginData: function(loginData, users, passwords) {
			var list = [];
			for (var i=0; i<loginData.length; i++) {
				if (this._userEntryMatches(loginData[i], null, users, passwords, null, null, true)) {
					list.push(loginData[i]);
				}
			}

			return list;
		},

		// returns all users in the same domain as hostname and matching user/passes
		findAllUsers: function(hostname, users, passwords) {
			if (users && !smIsArray(users)) {
				users = [users];
			}
			if (passwords && !smIsArray(passwords)) {
				passwords = [passwords];
			}
			var userDomain = smUtils.getRootDomain(hostname);
			var loginData = this._readDomainLoginData(userDomain);
			
			if (users || passwords) {
				return this._findAllUsersInLoginData(loginData, users, passwords);
			}
			
			return loginData;
		},
		
		_addUserToLoginData: function(loginData, hostname, users, passwords, formActionHostURL, httpRealm) {
			var userEntry = {
				hostname: hostname,
				users: users,
				passwords: passwords,
				formActionHostURL: formActionHostURL,
				httpRealm: httpRealm,
				created: new Date(),
			};

			loginData.push(userEntry);
			
			return userEntry;
		},
		
		addUser: function(hostname, users, passwords, formActionHostURL, httpRealm) {
			if (!smIsArray(users)) {
				users = [users];
			}
			if (!smIsArray(passwords)) {
				passwords = [passwords];
			}
			var userDomain = smUtils.getRootDomain(hostname);
			var loginData = this._readDomainLoginData(userDomain);
			
			var userEntry = this._findUserInLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);
			if (userEntry) {
				//throw "smDataStore.passwordmanager.addUser: Error: user already exist";
				smLog("smDataStore.passwordmanager.addUser: error user already exists: " + [hostname, users, passwords, formActionHostURL, httpRealm].join(','));
				return;
			}

			userEntry = this._addUserToLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);

			this._writeDomainLoginData(userDomain, loginData);
		},

		_removeUserInLoginData: function(loginData, hostname, users, passwords, formActionHostURL, httpRealm) {
			var deletedCount = 0;
			for (var i=0; i<loginData.length; i++) {
				if (this._userEntryMatches(loginData[i], hostname, users, passwords, formActionHostURL, httpRealm)) {
					loginData.splice(i,1);
					deletedCount++;
				}
			}
			return deletedCount;
		},

		// TODO: review
		removeUser: function(hostname, users, passwords, formActionHostURL, httpRealm) {
			if (!smIsArray(users)) {
				users = [users];
			}
			if (!smIsArray(passwords)) {
				passwords = [passwords];
			}
			var userDomain = smUtils.getRootDomain(hostname);
			var loginData = this._readDomainLoginData(userDomain);
			
			/*
			var userEntry = this._findUserInLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);
			if (!userEntry) {
				smlog("smDataStore.passwordmanager.addUser: error no such user: " + [hostname, users, passwords, formActionHostURL, httpRealm].join(','));
				return;
			}
			*/
			var deletedCount = this._removeUserInLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);
			if (deletedCount == 0) {
				//throw "smDataStore.passwordmanager.addUser: Error: user already exist";
				smLog("smDataStore.passwordmanager.removeUser: error no such user: " + [hostname, users, passwords, formActionHostURL, httpRealm].join(','));
				return;
			}

			//smUtils.inspectObject(loginData);

			this._writeDomainLoginData(userDomain, loginData);
		},
		
 		// returns all users in the same domain as hostname and matching user/passes
 		removeAllUsers: function(hostname, users, passwords) {
			if (!smIsArray(users)) {
				users = [users];
			}
			if (!smIsArray(passwords)) {
				passwords = [passwords];
			}
			var userDomain = smUtils.getRootDomain(hostname);
			var loginData = this._readDomainLoginData(userDomain);
			
			var list = this._findAllUsersInLoginData(loginData, users, passwords);
			
			var deletedCount = 0;
			for (var i=0; i < list.length; i++) {
				deletedCount += this._removeUserInLoginData(loginData, list[i].hostname, list[i].users, list[i].passwords, list[i].formActionHostURL, list[i].httpRealm);
			}
			if (deletedCount == 0) {
				smLog("smDataStore.passwordmanager.removeAllUsers: error no such users: " + [users, passwords].join(','));
				return;
			}

			this._writeDomainLoginData(userDomain, loginData);
		},
		
		ledgerDomainEvent: function(eventType, domain, eventObj) {
			var loginType;
			if (eventType == this.dataStore.ledger.DOMAIN_EVENT_TYPE.LOGIN) {
				loginType = 'successful';

				//// store last login on the whole domain
				this.dataStore.ledger.setDomainState(domain,
							this.dataStore.ledger.DOMAIN_STATE_TYPE.LAST_LOGIN, 
							smUtils.mergeAssocArray({loginEventID: eventObj.id, date: eventObj.date}, eventObj.data)
						);

			} else if (eventType == this.dataStore.ledger.DOMAIN_EVENT_TYPE.LOGIN_UNSUCCESSFUL) {
				loginType = 'unsuccessful';
			} else {
				return;
			}
			
			//// store login meta data
			var url = eventObj.data['url'];
			var userDomain = smUtils.getRootDomain(url);
			var hostname = smUtils.getHostURL(url);
			var formHash = eventObj.data['formHash'];
			var httpRealm = eventObj.data['httpRealm'];
			//var partial = eventObj.data['partial'];
			
			var formActionHostURL = eventObj.data['formActionHostURL'];
			var formActionHostURLnonUnified = eventObj.data['formActionHostURLnonUnified'];

			var users = [];
			eventObj.data['users'].forEach(function(u) {
					users.push(u.value);
				});
			//users = users.sort();

			var passwords =[]
			eventObj.data['passwords'].forEach(function(p) {
					passwords.push(p.value);
				});
			//passwords = passwords.sort();

			var loginData = this._readDomainLoginData(userDomain);

			var userEntry = this._findUserInLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);
			var count = 0;
			if (!userEntry) {
				userEntry = this._addUserToLoginData(loginData, hostname, users, passwords, formActionHostURL, httpRealm);
				userEntry.created = eventObj.date;
				count = 0;
			} else {
				if (userEntry[loginType] && typeof(userEntry[loginType].count) != 'undefined') {
					count = userEntry[loginType].count;
				}
			}
			
			count++;
			userEntry.isLastLoginSuccess = (eventType == this.dataStore.ledger.DOMAIN_EVENT_TYPE.LOGIN); 
			userEntry.lastDate = eventObj.date; 
			userEntry[loginType] = {
						loginEventID: eventObj.id,
						lastDate: eventObj.date,
						url: url,
						formHash: formHash,
						count: count
					};
			this._writeDomainLoginData(userDomain, loginData);
			
			// add account to FF password store as well - to match hostname
                        // Note: This code executes on login form scraping - user may have not acknowledged to remember account
			if (eventType == this.dataStore.ledger.DOMAIN_EVENT_TYPE.LOGIN 
			  && users.length==1 && passwords.length==1) {
				var username = users[0];
				var password = passwords[0];
                                try {
                                        var users = this.dataStore.userManager.getUsersFromHost(userDomain, username, password);
                                        var isAccountPerDomainExisting = false;
                                        for(var i=0; i<users.length; i++) {
                                                if (users[i].user==username) {
                                                        isAccountPerDomainExisting = true;
                                                        break;                                   
                                                }
                                        }
                                        // See above note: do not store account unless user has previously acknowledged to store it
                                        // 'exact' duplicates are filtered later
                                        if (isAccountPerDomainExisting) {
                                                var user = this.dataStore.userManager.createUser(hostname, username, password, httpRealm, formActionHostURLnonUnified);
                                        }
                                } catch (e) {
                                        // ignore errors like account already exists
                                }
			}
		}
					
	},
	/*
	 
	,
	
	test: function() {
		//this.file.test();
		this.ledger.test();

	}
	*/

	/* nsISupports */
	QueryInterface : function handler_QI(iid) {
		if (iid.equals(Components.interfaces.nsISupports))
			return this;

		throw Components.results.NS_ERROR_NO_INTERFACE;
	}
		
};

/***********************************************************
USAGE

smDataStore = Components.classes["@sxip.com/sxipper/datastore;1"].
                          getService().wrappedJSObject;
                        //!  createInstance(Components.interfaces.nsIsmDataStore).wrappedJSObject;

***********************************************************/

var singleton;

var smDataStoreFactory = {
  createInstance: function (aOuter, aIID) {
    if (aOuter != null) {
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    }
    if (!singleton) {
		singleton = new smIDataStore();
    }
    return singleton.QueryInterface(aIID);
  }
};

/***********************************************************
module definition (xpcom registration)
***********************************************************/
var smDataStoreModule = {
  _firstTime: true,

  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME,
        CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
  	try {
	  	if (singleton) {
	  		// TODO: clean up
	  	}
  	} catch (e) {
  		dump("error unloading smDataStore");
  	}
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);
  },

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return smDataStoreFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};


/***********************************************************
module initialization

When the application registers the component, this function
is called.
***********************************************************/
function NSGetModule(aCompMgr, aFileSpec) { return smDataStoreModule; }

